local print = function(msg) if msg then DEFAULT_CHAT_FRAME:AddMessage(msg) end end
--local pdebug = function(msg) if msg then table.insert(UABDebug, msg) end end
--local pdebug = function(msg) print(msg) end
local pdebug = function(msg) end

--[[
-----------------------------------------------------------------------------------
Class: UnitActionBars.layouts.uaboval
Static Methods:
Instance Methods:
-----------------------------------------------------------------------------------
--]]
UnitActionBars.layouts.uaboval = UnitActionBars.aceOO.Class(UnitActionBars.layoutManager.ILayout, UnitActionBars.aceEvent)

function UnitActionBars.layouts.uaboval.GetLayoutName()
    return "UABOval"
end

function UnitActionBars.layouts.uaboval.GetDisplayName()
    return "Oval"
end

function UnitActionBars.layouts.uaboval.GetConfigTemplate()
    return "UAB_OvalLayoutOptionsTemplate"
end
     
function UnitActionBars.layouts.uaboval.prototype:init(configFrame, sampleFrame, settings)
    UnitActionBars.layouts.uaboval.super.prototype.init(self)

    self:RegisterEvent("UnitActionBars_CreateGroupUnit", "CreateGroupUnit")

    self.configFrame = configFrame
    self.sampleFrame = sampleFrame
    configFrame.layout = self
    
    self.controllers = {}
    self.settings = UnitActionBars.layouts.uaboval.settings:new(configFrame, settings)
    self.sampleController = UnitActionBars.layouts.uaboval.controller:new(self.settings, "sample", 1, sampleFrame, 1)

    configFrame:SetScript("OnShow", function(self) self.layout:ShowSample() end)
    configFrame:SetScript("OnHide", function(self) self.layout:HideSample() end)
end

function UnitActionBars.layouts.uaboval.prototype:ShowSample()
    self.sampleController:EnableActionSet()
end

function UnitActionBars.layouts.uaboval.prototype:HideSample()
    self.sampleController:DisableActionSet()
end

function UnitActionBars.layouts.uaboval.prototype:CreateGroupUnit(name, id, header)
    self.controllers[name..id] = UnitActionBars.layouts.uaboval.controller:new(self.settings, name, id, header)
end

function UnitActionBars.layouts.uaboval.prototype:GetSettings()
    self.settings:GetValues()
end

function UnitActionBars.layouts.uaboval.prototype:DefaultSettings(bypassNotification)
    self.settings:Defaults(bypassNotification)
end

--[[
-----------------------------------------------------------------------------------
Class: UnitActionBars.layouts.uaboval.controller
Static Methods:
Instance Methods:
-----------------------------------------------------------------------------------
--]]
UnitActionBars.layouts.uaboval.controller = UnitActionBars.aceOO.Class(UnitActionBars.aceEvent)

--[[
This table defines the placement order for the buttons assuming
a specific starting point.  Each row is positioned, then the next
row.  Iterate over the button order positions until all have been
used, then go to the next row.  This means position 6 buttons in
the first row, 12 in the second, and 6 in the third.

To use this table when not starting at the specific starting point
use the buttonOrderSequence table to find the first index for each
row for a specific starting point.
--]]
UnitActionBars.layouts.uaboval.controller.rowButtonStart = { 1, 7, 19 }
UnitActionBars.layouts.uaboval.controller.rowButtons = { 6, 12, 6 }
UnitActionBars.layouts.uaboval.controller.buttonOrder = {
    [uab.constants.anchor.left] = { row1 = { "l2", "u2l1", "u2r1", "r2", "d2r1", "d2l1" },
          row2 = { "l4", "u2l3", "u4l2", "u4", "u4r2", "u2r3", "r4", "d2r3", "d4r2", "d4", "d4l2", "d2l3" },
          row3 = { "l6", "u2l5", "u2r5", "r6", "d2r5", "d2l5" }
    },
    [uab.constants.anchor.top] = { row1 = { "u2", "u1r2", "d1r2", "d2", "d1l2", "u1l2" },
          row2 = { "u4", "u3r2", "u2r4", "r4", "d2r4", "d3r2", "d4", "d3l2", "d2l4", "l4", "u2l4", "u3l2" },
          row3 = { "u6", "u5r2", "d5r2", "d6", "d5l2", "u5l2" }
    }
}

UnitActionBars.layouts.uaboval.controller.buttonOrderSequence = {
    [uab.constants.anchor.left] = {
          [uab.constants.placement.left] = { 1, 1, 1 },
          [uab.constants.placement.topleft] = { 2, 2, 2 },
          [uab.constants.placement.topright] = { 3, 4, 3 },
          [uab.constants.placement.right] = { 4, 7, 4 },
          [uab.constants.placement.bottomright] = { 5, 8, 5 },
          [uab.constants.placement.bottomleft] = { 6, 10, 6 }
    },
    [uab.constants.anchor.top] = {
          [uab.constants.placement.top] = { 1, 1, 1 },
          [uab.constants.placement.topright] = { 2, 2, 2 },
          [uab.constants.placement.bottomright] = { 3, 4, 3 },
          [uab.constants.placement.bottom] = { 4, 7, 4 },
          [uab.constants.placement.bottomleft] = { 5, 8, 5 },
          [uab.constants.placement.topleft] = { 6, 10, 6 }
    }
}
        
function UnitActionBars.layouts.uaboval.controller.prototype:init(settings, name, id, header, sample)
    UnitActionBars.layouts.uaboval.controller.super.prototype.init(self)
    
    self.id = id
    self.settings = settings
    self.sample = sample
    self.header = header
    self.groupname = name
    self.reposition = 1
    self.revisual = 1

    local actionsetname
    
    if sample then
        actionsetname = header:GetName().."UABOvalSample"
    
        self.header = CreateFrame("Frame", actionsetname.."Frame", header)
        self.header:Hide()
        self.header:SetHeight(header:GetWidth())
        self.header:SetWidth(header:GetHeight())
        self.header:SetPoint("CENTER")
    end
    
    self.actionset = UnitActionBars.actionsets.standard:GetActionSet(self.header, id, actionsetname)

    self:RegisterEvent("UnitActionBars_UABOval_SettingsChanged", "SettingsChanged")
    self:RegisterEvent("UnitActionBars_"..self.groupname.."_ChangeVisualActionEffect", "ChangeVisualEffect")
    self:RegisterEvent("UnitActionBars_UABOval_Enable", "EnableActionSet")
    self:RegisterEvent("UnitActionBars_UABOval_Disable", "DisableActionSet")
    
    if not sample then
        self:RegisterEvent("UnitActionBars_"..name.."_SetAction", "SetAction")
        self:RegisterEvent("UnitActionBars_"..name.."_RemoveAction", "RemoveAction")
    end

    self.actionset:SetInUse()
    if self.sample then
        self.header:Hide()
    end
end

function UnitActionBars.layouts.uaboval.controller.prototype:SettingsChanged(settings)
    self.settings = settings
    self.reposition = 1
    self:UpdatePositions()
end

function UnitActionBars.layouts.uaboval.controller.prototype:ChangeVisualEffect(id, effects)
    self.revisual = 1
    self:TriggerEvent("UnitActionBars_"..self.actionset:ActionSetName().."_ChangeVisualActionEffect"..id, effects)
end

function UnitActionBars.layouts.uaboval.controller.prototype:EnableActionSet()
    if not self.enabled then
        self.enabled = 1
        self.actionset:SetController(self)
        self.actionset:SetInUse(1)
        self.reposition = 1
        self.revisual = 1
        self:UpdateAll()
        
        if self.sample then
            self.header:Show()
        end
    end
end

function UnitActionBars.layouts.uaboval.controller.prototype:DisableActionSet()
    if self.enabled then
        self.enabled = nil    
        self.actionset:SetInUse()
        if self.sample then
            self.header:Hide()
        end
    end
end

function UnitActionBars.layouts.uaboval.controller.prototype:SetAction(number, action)
    self.reposition = 1
    self.revisual = 1
    self.actionset:SetAction(number, action)
end

function UnitActionBars.layouts.uaboval.controller.prototype:RemoveAction(number)
    if self.actionset:ActionButton(number) then
        self.reposition = 1
        self.revisual = 1
        self.actionset:RemoveAction(number)
    end
end

function UnitActionBars.layouts.uaboval.controller.prototype:UpdateAll()
    self:UpdatePositions()
    self:UpdateVisualEffects()
end

function UnitActionBars.layouts.uaboval.controller.prototype:UpdatePositions()
    if self.reposition and self.actionset:IsVisible() then
        self.reposition = nil
        
        -- repositioning logic goes here
        self.actionset:ClearPositions()        

        self:PositionRow(1)
        self:PositionRow(2)
        self:PositionRow(3)
    end
end

function UnitActionBars.layouts.uaboval.controller.prototype:PositionRow(row)
    local order = UnitActionBars.layouts.uaboval.controller.buttonOrder[self.settings.anchor]["row"..row]
    local sequence = UnitActionBars.layouts.uaboval.controller.buttonOrderSequence[self.settings.anchor]
    local startpos = self.settings.startLocation
    local buttonIdx = UnitActionBars.layouts.uaboval.controller.rowButtonStart[row]

    if self.settings.growClockwise == "true" then
        for idx = sequence[startpos][row], UnitActionBars.layouts.uaboval.controller.rowButtons[row], 1 do
            self:PositionButton(order[idx], buttonIdx, self.actionset:ActionButton(buttonIdx), self.actionset:CancelButton())
            buttonIdx = buttonIdx + 1
        end
        
        local endpos = sequence[startpos][row] - 1
        
        for idx = 1, endpos, 1 do
            self:PositionButton(order[idx], buttonIdx, self.actionset:ActionButton(buttonIdx), self.actionset:CancelButton())
            buttonIdx = buttonIdx + 1
        end
    else
        for idx = sequence[startpos][row], 1, -1 do
            self:PositionButton(order[idx], buttonIdx, self.actionset:ActionButton(buttonIdx), self.actionset:CancelButton())
            buttonIdx = buttonIdx + 1
        end
        
        local endpos = sequence[startpos][row] + 1
        
        for idx = UnitActionBars.layouts.uaboval.controller.rowButtons[row], endpos, -1 do
            self:PositionButton(order[idx], buttonIdx, self.actionset:ActionButton(buttonIdx), self.actionset:CancelButton())
            buttonIdx = buttonIdx + 1
        end
    end
end

function UnitActionBars.layouts.uaboval.controller.prototype:PositionButton(orderCode, btnidx, btn, center)
    if not btn then return end
    
    local vMod = 0
    local hMod = 0

    tmpCode = orderCode        
    if string.sub(orderCode, 1, 1) == "u" then
        vMod = tonumber(string.sub(orderCode, 2, 2)) * 0.5
        tmpCode = string.sub(orderCode, 3)
    end

    if string.sub(tmpCode, 1, 1) == "d" then
        vMod = tonumber(string.sub(tmpCode, 2, 2)) * -0.5
        tmpCode = string.sub(tmpCode, 3)
    end

    if string.sub(tmpCode, 1, 1) == "r" then
        hMod = tonumber(string.sub(tmpCode, 2, 2)) * 0.5
        tmpCode = string.sub(tmpCode, 3)
    end

    if string.sub(tmpCode, 1, 1) == "l" then
        hMod = tonumber(string.sub(tmpCode, 2, 2)) * -0.5
        tmpCode = string.sub(tmpCode, 3)
    end

    local x = (UnitActionBars.actionbuttons.standard.buttonWidth + 1) * hMod
    local y = (UnitActionBars.actionbuttons.standard.buttonHeight + 1) * vMod
    
    btn:SetPoint("CENTER", center, "CENTER", x, y)
end

function UnitActionBars.layouts.uaboval.controller.prototype:UpdateVisualEffects()
    if self.revisual and self.actionset:IsVisible() then
        self.revisual = nil
    
        -- apply all the visual properties based on user selections (low alpha, highlight border, blink)
        -- for each button control in the actionset

        --self.actionset:UpdateVisualEffects()        
    end
end

--[[
-----------------------------------------------------------------------------------
Class: UnitActionBars.layouts.uaboval.settings
Static Methods:
Instance Methods:
-----------------------------------------------------------------------------------
--]]
UnitActionBars.layouts.uaboval.settings = UnitActionBars.aceOO.Class(UnitActionBars.aceEvent)

function UnitActionBars.layouts.uaboval.settings.AnchorDisplayValue(value)
    if value == uab.constants.anchor.top then
        return "Top & Bottom"
    else
        return "Left & Right"
    end
end

function UnitActionBars.layouts.uaboval.settings.StartLocationDisplayValue(value)
    if value == uab.constants.placement.top then
        return "Top"
    elseif value == uab.constants.placement.topright then
        return "Top Right"
    elseif value == uab.constants.placement.topleft then
        return "Top Left"
    elseif value == uab.constants.placement.bottom then
        return "Bottom"
    elseif value == uab.constants.placement.bottomright then
        return "Bottom Right"
    elseif value == uab.constants.placement.bottomleft then
        return "Bottom Left"
    elseif value == uab.constants.placement.left then
        return "Left"
    else
        return "Right"
    end
end

function UnitActionBars.layouts.uaboval.settings.prototype:init(configFrame, settings)
    UnitActionBars.layouts.uaboval.settings.super.prototype.init(self)
    
    self.frame = configFrame

    self.growClockwise = (settings and settings.growClockwise) or "true"
    self.anchor = (settings and settings.anchor) or uab.constants.anchor.left
    self.startLocation = (settings and settings.startLocation) or uab.constants.placement.topleft

    self.growDirectionFrame = getglobal(self.frame:GetName().."GrowDirection")
    self.growDirectionFrame.settings = self
	
    getglobal(self.growDirectionFrame:GetName().."Text"):SetText("Add buttons clockwise")
	self.growDirectionFrame:SetScript("OnClick", function(self, button, down) self.settings:GrowDirectionButtonClicked(self) end)

    self.startLocationFrame = getglobal(self.frame:GetName().."StartLocation")
    self.startLocationFrame.settings = self

    UnitActionBars.layouts.uaboval.settings.InsideInit = 1
    UnitActionBars.layouts.uaboval.settings.InitDropdown = self
    
    UIDropDownMenu_Initialize(self.startLocationFrame,
        function()
            if UnitActionBars.layouts.uaboval.settings.InsideInit then
                UnitActionBars.layouts.uaboval.settings.InitDropdown:StartLocationListInitialize()
            else
                this:GetParent().settings:StartLocationListInitialize()
            end
        end)
	UIDropDownMenu_SetWidth(110, self.startLocationFrame)
	UIDropDownMenu_SetButtonWidth(24, self.startLocationFrame)
	UIDropDownMenu_JustifyText("LEFT", self.startLocationFrame)

    self.anchorFrame = getglobal(self.frame:GetName().."Anchor")
    self.anchorFrame.settings = self
	
    UnitActionBars.layouts.uaboval.settings.InsideInit = 1
    UnitActionBars.layouts.uaboval.settings.InitDropdown = self
    
    UIDropDownMenu_Initialize(self.anchorFrame,
        function()
            if UnitActionBars.layouts.uaboval.settings.InsideInit then
                UnitActionBars.layouts.uaboval.settings.InitDropdown:AnchorListInitialize()
            else
                this:GetParent().settings:AnchorListInitialize()
            end
        end)
	UIDropDownMenu_SetWidth(110, self.anchorFrame)
	UIDropDownMenu_SetButtonWidth(24, self.anchorFrame)
	UIDropDownMenu_JustifyText("LEFT", self.anchorFrame)

    UnitActionBars.layouts.uaboval.settings.InsideInit = nil
    UnitActionBars.layouts.uaboval.settings.InitDropdown = nil
    
	self:SetGrowClockwise(self.growClockwise)
	self:SetStartLocation(self.startLocation)
	self:SetAnchor(self.anchor)
    
    self:RegisterEvent("UnitActionBars_ProfileChanged", "ProfileChanged")    
    self:TriggerEvent("UnitActionBars_UABOval_SettingsChanged", self)
end

function UnitActionBars.layouts.uaboval.settings.prototype:ProfileChanged()
    local settings = UnitActionBars.db.profile.layoutSettings.UABOval
    
    self.growClockwise = (settings and settings.growClockwise) or "true"
    self.anchor = (settings and settings.anchor) or uab.constants.anchor.left
    self.startLocation = (settings and settings.startLocation) or uab.constants.placement.topleft
    
	self:SetGrowClockwise(self.growClockwise)
	self:SetStartLocation(self.startLocation)
	self:SetAnchor(self.anchor)
end

function UnitActionBars.layouts.uaboval.settings.prototype:Defaults(bypassNotification)
    self.growClockwise = "true"
    self.anchor = uab.constants.anchor.left
    self.startLocation = uab.constants.placement.topleft

    UnitActionBars.db.profile.layoutSettings.UABOval = self:GetValues()
    
    if not bypassNotification then
        self:TriggerEvent("UnitActionBars_UABOval_SettingsChanged", self)
    end
end

function UnitActionBars.layouts.uaboval.settings.prototype:StartLocationListInitialize()
    local index = 1
    local selected = nil

    local options = {}

    table.insert(options, {	text = "Top Left",
                            func = UAB_DropDownSelect,
                            arg1 = self,
                            arg2 = { func = "SelectStartLocation", arg1 = uab.constants.placement.topleft }})
    table.insert(options, {	text = "Top",
                            func = UAB_DropDownSelect,
                            arg1 = self,
                            arg2 = { func = "SelectStartLocation", arg1 = uab.constants.placement.top }})
    table.insert(options, {	text = "Top Right",
                            func = UAB_DropDownSelect,
                            arg1 = self,
                            arg2 = { func = "SelectStartLocation", arg1 = uab.constants.placement.topright }})
    table.insert(options, {	text = "Right",
                            func = UAB_DropDownSelect,
                            arg1 = self,
                            arg2 = { func = "SelectStartLocation", arg1 = uab.constants.placement.right }})
    table.insert(options, {	text = "Bottom Right",
                            func = UAB_DropDownSelect,
                            arg1 = self,
                            arg2 = { func = "SelectStartLocation", arg1 = uab.constants.placement.bottomright }})
    table.insert(options, {	text = "Bottom",
                            func = UAB_DropDownSelect,
                            arg1 = self,
                            arg2 = { func = "SelectStartLocation", arg1 = uab.constants.placement.bottom }})
    table.insert(options, {	text = "Bottom Left",
                            func = UAB_DropDownSelect,
                            arg1 = self,
                            arg2 = { func = "SelectStartLocation", arg1 = uab.constants.placement.bottomleft }})
    table.insert(options, {	text = "Left",
                            func = UAB_DropDownSelect,
                            arg1 = self,
                            arg2 = { func = "SelectStartLocation", arg1 = uab.constants.placement.left }})

    for index, data in pairs(options) do
        UIDropDownMenu_AddButton(data)
        if self.startLocation == data.arg2.arg1 then
            selected = index
        end
    end

    UIDropDownMenu_SetSelectedID(self.startLocationFrame, selected)
end

function UnitActionBars.layouts.uaboval.settings.prototype:SelectStartLocation(selector, selection)
    UIDropDownMenu_SetSelectedID(self.startLocationFrame, selector:GetID())
    self:SetStartLocation(selection)
end

function UnitActionBars.layouts.uaboval.settings.prototype:SetStartLocation(value)
    local notify = (value ~= self.startLocation)
        
    self.startLocation = value
    UIDropDownMenu_SetText(UnitActionBars.layouts.uaboval.settings.StartLocationDisplayValue(value), self.startLocationFrame)
    
    UnitActionBars.db.profile.layoutSettings.UABOval = self:GetValues()
    
    if notify then
        self:TriggerEvent("UnitActionBars_UABOval_SettingsChanged", self)
    end        
end

function UnitActionBars.layouts.uaboval.settings.prototype:AnchorListInitialize()
    local index = 1
    local selected = nil

    local anchorOptions = {}

    table.insert(anchorOptions, {	text = "Top & Bottom",
                                    func = UAB_DropDownSelect,
                                    arg1 = self,
                                    arg2 = { func = "SelectAnchor", arg1 = uab.constants.anchor.top }})
    table.insert(anchorOptions, {	text = "Left & Right",
                                    func = UAB_DropDownSelect,
                                    arg1 = self,
                                    arg2 = { func = "SelectAnchor", arg1 = uab.constants.anchor.left }})

    for index, data in pairs(anchorOptions) do
        UIDropDownMenu_AddButton(data)
        if self.anchor == data.arg2.arg1 then
            selected = index
        end
    end

    UIDropDownMenu_SetSelectedID(self.anchorFrame, selected)
end

function UnitActionBars.layouts.uaboval.settings.prototype:SelectAnchor(selector, selection)
    UIDropDownMenu_SetSelectedID(self.anchorFrame, selector:GetID())
    self:SetAnchor(selection)
end

function UnitActionBars.layouts.uaboval.settings.prototype:SetAnchor(value)
    local notify = (value ~= self.anchor)
    
    self.anchor = value
    
    UIDropDownMenu_SetText(UnitActionBars.layouts.uaboval.settings.AnchorDisplayValue(value), self.anchorFrame)
    
    if value == uab.constants.anchor.top then
        if self.startLocation == uab.constants.placement.left or self.startLocation == uab.constants.placement.right then
            self:SetStartLocation(uab.constants.placement.top..self.startLocation)
        end
    elseif value == uab.constants.anchor.left then
        if self.startLocation == uab.constants.placement.top or self.startLocation == uab.constants.placement.bottom then
            self:SetStartLocation(self.startLocation..uab.constants.placement.left)
        end
    end
    
    UnitActionBars.db.profile.layoutSettings.UABOval = self:GetValues()
    
    if notify then
        self:TriggerEvent("UnitActionBars_UABOval_SettingsChanged", self)
    end        
end

function UnitActionBars.layouts.uaboval.settings.prototype:GrowDirectionButtonClicked(frame)
    self:SetGrowClockwise(frame:GetChecked())
end

function UnitActionBars.layouts.uaboval.settings.prototype:SetGrowClockwise(value)
    local newval
    
    if value and (tonumber(value) == 1 or value == "true") then
        newval = "true"
    else
        newval = "false"
    end
    
    local notify = (newval ~= self.growClockwise)
        
    self.growClockwise = newval
    self.growDirectionFrame:SetChecked(self.growClockwise == "true")
    
    UnitActionBars.db.profile.layoutSettings.UABOval = self:GetValues()
    
    if notify then
        self:TriggerEvent("UnitActionBars_UABOval_SettingsChanged", self)
    end        
end

function UnitActionBars.layouts.uaboval.settings.prototype:GetValues()
    local values = {}
    
    values.growClockwise = self.growClockwise
    values.anchor = self.anchor
    values.startLocation = self.startLocation
    
    return values
end

UnitActionBars.layoutManager:Register( UnitActionBars.layouts.uaboval )
